import { Pushable } from 'it-pushable';
import { Libp2p, Libp2pInit } from '@cerc-io/libp2p';
import type { Stream as P2PStream, Connection } from '@libp2p/interface-connection';
import type { PeerInfo } from '@libp2p/interface-peer-info';
import type { Message } from '@libp2p/interface-pubsub';
import type { PeerId } from '@libp2p/interface-peer-id';
import { Multiaddr } from '@multiformats/multiaddr';
import { PrometheusMetrics } from '@cerc-io/prometheus-metrics';
import { PeerHearbeatChecker } from './peer-heartbeat-checker.js';
import { PubsubType } from './utils/index.js';
import { DebugPeerInfo, PeerConnectionInfo, PeerSelfInfo } from './types/debug-info.js';
export interface PeerIdObj {
    id: string;
    privKey: string;
    pubKey: string;
}
export interface PeerInitConfig {
    pingInterval?: number;
    pingTimeout?: number;
    maxRelayConnections?: number;
    relayRedialInterval?: number;
    denyMultiaddrs?: string[];
    maxConnections?: number;
    minConnections?: number;
    dialTimeout?: number;
    pubsub?: PubsubType;
    directPeers?: string[];
    enableDebugInfo?: boolean;
    transports?: Libp2pInit['transports'];
    listenMultiaddrs?: string[];
    peerDiscovery?: Libp2pInit['peerDiscovery'];
}
export declare class Peer {
    _node?: Libp2p;
    _peerHeartbeatChecker?: PeerHearbeatChecker;
    _webRTCSignallingEnabled: boolean;
    _relayNodeMultiaddr: Multiaddr;
    _numRelayConnections: number;
    _relayRedialInterval?: number;
    _maxRelayConnections?: number;
    _denyMultiaddrs?: string[];
    _debugInfoEnabled?: boolean;
    _peerStreamMap: Map<string, Pushable<any>>;
    _messageHandlers: Array<(peerId: PeerId, message: any) => void>;
    _topicHandlers: Map<string, Array<(peerId: PeerId, data: any) => void>>;
    _metrics: PrometheusMetrics;
    constructor(relayNodeURL: string, nodejs?: boolean);
    get peerId(): PeerId | undefined;
    get node(): Libp2p | undefined;
    get relayNodeMultiaddr(): Multiaddr;
    get metrics(): PrometheusMetrics;
    init(initOptions: PeerInitConfig, peerIdObj?: PeerIdObj): Promise<void>;
    close(): Promise<void>;
    getInfo(): Promise<DebugPeerInfo>;
    getPeerSelfInfo(): PeerSelfInfo;
    getPeerConnectionsInfo(): PeerConnectionInfo[];
    broadcastMessage(message: any): void;
    floodMessage(topic: string, msg: any): Promise<void>;
    requestPeerInfo(): Promise<void>;
    subscribeMessage(handler: (peerId: PeerId, message: any) => void): () => void;
    subscribeTopic(topic: string, handler: (peerId: PeerId, data: any) => void): () => void;
    subscribeDebugInfo(handler: (peerId: PeerId, data: any) => void): () => void;
    isRelayPeerMultiaddr(multiaddrString: string): boolean;
    isPrimaryRelay(multiaddrString: string): boolean;
    getLatencyData(peerId: PeerId): Array<number>;
    _handleChangeProtocols({ peerId, protocols }: {
        peerId: PeerId;
        protocols: string[];
    }): Promise<void>;
    _dialRelay(redialInterval?: number): Promise<void>;
    _handleDiscovery(peer: PeerInfo, maxRelayConnections?: number): void;
    _handleConnect(connection: Connection, maxRelayConnections?: number): Promise<void>;
    _createProtocolStream(connection: Connection, protocol: string): Promise<void>;
    _handleDeadConnections(remotePeerId: PeerId): Promise<void>;
    _handleDisconnect(connection: Connection): Promise<void>;
    _connectPeer(peer: PeerInfo): Promise<void>;
    _handleStream(peerId: PeerId, stream: P2PStream): void;
    _handlePubSubMessage(msg: Message): void;
    _registerDebugInfoRequestHandler(): void;
}
export declare function createPeerId(): Promise<PeerIdObj>;
